"use strict"
const async = require ("async");
let $o = require ("/Users/dimas/objectum/node/objectum");
let code = "region_test";
let storage;
function loadRecs ({session}, cb) {
	let cRecs;
	async.series ([
		function (cb) {
			$o.projects.loadConfig ({
				code: code, success: function () {
					cb ();
				}, failure: function (err) {
					cb (err);
				}
			});
		},
		function (cb) {
			let connection = $o.config.storages [code];
			storage = new $o.Storage ({
				code: code, connection: connection, success: function () {
					cb ();
				}
			});
		},
		function (cb) {
			storage.query ({session, sql: `select distinct fid from tclass where fid >= 1000 and fend_id = 2147483647`}, function (err, recs) {
				cRecs = recs;
				cb (err);
			});
		}
	], function (err) {
		cb (err, {cRecs});
	});
};
function convert ({session}, cb) {
	let cRecs, caRecs;
	async.series ([
		function (cb) {
			loadRecs ({session}, function (err, opts) {
				cRecs = opts.cRecs;
				caRecs = opts.caRecs;
				cb (err);
			});
		},
		function (cb) {
			async.eachSeries (cRecs, function (rec, cb) {
				storage.client.isTableExists ({session: session, table: `tobject_${rec.fid}`, success: function (result) {
					if (!result) {
						return cb ();
					};
					console.log (`Removing tobject_${rec.fid}`);
					storage.query ({session: session, sql: `drop table $schema_prefix$tobject_${rec.fid}`}, cb);
				}});
			}, cb);
		},
		function (cb) {
			async.eachSeries (cRecs, function (rec, cb) {
				storage.client.isTableExists ({session: session, table: `tobject_attr_${rec.fid}`, success: function (result) {
					if (!result) {
						return cb ();
					};
					console.log (`Removing tobject_attr_${rec.fid}`);
					storage.query ({session: session, sql: `drop table $schema_prefix$tobject_attr_${rec.fid}`}, cb);
				}});
			}, cb);
		},
		function (cb) {
			async.eachSeries (cRecs, function (rec, cb) {
				let sql = `
					create table $schema_prefix$tobject_attr_${rec.fid} (
						fid $tid_object_attr$ not null,
						fobject_id $tnumber$,
						fclass_attr_id $tnumber$,
						fstring $tstring_value$,
						fnumber $tnumber_value$,
						ftime $ttimestamp$,
						fstart_id $tnumber$,
						fend_id $tnumber$,
						fschema_id $tnumber$,
						frecord_id $tnumber$
					) $tablespace$;
				`;
				console.log (`Creating tobject_attr_${rec.fid}`);
				storage.query ({session: session, sql: sql}, cb);
			}, cb);
		},
		function (cb) {
			async.eachSeries (cRecs, function (rec, cb) {
				let sql = `
					create index tobject_attr_${rec.fid}_fid on $schema_prefix$tobject_attr_${rec.fid} (fid) $tablespace$;
					create index tobject_attr_${rec.fid}_fclass_attr_id on $schema_prefix$tobject_attr_${rec.fid} (fclass_attr_id) $tablespace$;
					create index tobject_attr_${rec.fid}_fend_id on $schema_prefix$tobject_attr_${rec.fid} (fend_id) $tablespace$;
					create index tobject_attr_${rec.fid}_fnumber on $schema_prefix$tobject_attr_${rec.fid} (fnumber) $tablespace$;
					create index tobject_attr_${rec.fid}_fobject_id on $schema_prefix$tobject_attr_${rec.fid} (fobject_id) $tablespace$;
					create index tobject_attr_${rec.fid}_ftime on $schema_prefix$tobject_attr_${rec.fid} (ftime) $tablespace$;
					create index tobject_attr_${rec.fid}_fstart_id on $schema_prefix$tobject_attr_${rec.fid} (fstart_id) $tablespace$;
					create index tobject_attr_${rec.fid}_fstring on $schema_prefix$tobject_attr_${rec.fid} ($tobject_attr_fstring$);
					create index tobject_attr_${rec.fid}_fschema_id on $schema_prefix$tobject_attr_${rec.fid} (fschema_id) $tablespace$;
					create index tobject_attr_${rec.fid}_frecord_id on $schema_prefix$tobject_attr_${rec.fid} (frecord_id) $tablespace$;
				`;
				console.log (`Creating indexes tobject_attr_${rec.fid}`);
				storage.query ({session: session, sql: sql}, cb);
			}, cb);
		},
		function (cb) {
			storage.query ({session: session, sql: `alter table tobject_attr drop column fclass_id`}, cb);
		},
		function (cb) {
			storage.query ({session: session, sql: `alter table tobject_attr add fclass_id bigint`}, cb);
		},
		function (cb) {
			let i = 0;
			async.eachSeries (cRecs, function (rec, cb) {
				let ca = storage.getClass (rec.fid).attrs;
				ca = _.map (ca, function (rec) {
					return rec.get ("fid");
				});
				if (!ca.length) {
					return cb ();
				};
				let sql = `
					update tobject_attr set fclass_id = ${rec.fid} where fclass_attr_id in (${ca.join ()})
				`;
				console.log (`Updating tobject_attr fclass_id = ${rec.fid} (${++ i} of ${cRecs.length})`);
				storage.query ({session: session, sql: sql}, cb);
			}, cb);
		},
		function (cb) {
			async.eachSeries (cRecs, function (rec, cb) {
				let sql = `
					insert into tobject_attr_${rec.fid} (
						fid,
						fobject_id,
						fclass_attr_id,
						fstring,
						fnumber,
						ftime,
						fstart_id,
						fend_id,
						fschema_id,
						frecord_id
					)
					select
						fid,
						fobject_id,
						fclass_attr_id,
						fstring,
						fnumber,
						ftime,
						fstart_id,
						fend_id,
						fschema_id,
						frecord_id
					from tobject_attr
					where fclass_id = ${rec.fid}
				`;
				console.log (`Inserting to tobject_attr_${rec.fid} (${++ i} of ${cRecs.length})`);
				storage.query ({session: session, sql: sql}, cb);
			}, cb);
		}
	], function (err) {
		cb (err);
	});
};
let session = {
	id: "converter",
	username: "admin",
	userId: null
};
/*
convert ({session}, function (err) {
	console.log (err || "ok");
	process.exit (1);
});*/

function f1 (opts, cb) {
	console.log ("f1", opts);
	cb ();
};
function f2 (opts, cb) {
	console.log ("f2", opts);
	cb ();
};
async.series ([
	_.bind (f1, this, {}),
	_.bind (f2, this, {a: 1})
], function (err) {
	console.log (err || "ok");
	process.exit (1);
});
