"use strict"
class Objectum {
	constructor (config) {
		var me = this;
		if (config) {
			config = config.config || config;
		};
		if (!config) {
			config = require ("./config");
		};
		global.config = config;

		global._ = require ("lodash");
		_.findWhere = _.findLast;
		_.where = _.filter;
		// default values
		if (!config.auth) {
			config.auth = {multi: true};
		};
		config.backlog = config.backlog || 10000;
		if (!config.caching) {
			config.caching = {enabled: 0};
		};
		if (!config.admin) {
			config.admin = {ip: ["127.0.0.1"]};
		};
		if (!config.session) {
			config.session = {
				timeoutInterval: 120 * 1000,
				gcInterval: 300 * 1000
			};
		};
		if (!config.news) {
			config.news = {
				pollingInterval: 5000,
				gcInterval: 300000
			};
		};
		config.wwwRoot = __dirname + "/../www";
		if (_.isArray (config.storages) && config.projectsDir) {
			let storages = {};
			_.each (config.storages, function (code) {
				storages [code] = require (config.projectsDir + "/" + code + "/config.json");
			});
			config.storages = storages;
		};
		// global variables
		global.async = require ("async");
		global.pg = require ("pg");
		pg.defaults.poolSize = 0;
		pg.defaults.poolIdleTimeout = 120000;
		pg.defaults.reapIntervalMillis = 60000;
		global.util = require ("util");
		global.fs = require ("fs");
		global.http = require ("http");
		global.url = require ("url");
		if (config.redis && config.redis.enabled) {
			global.redis = require ("redis");
		};
		global.express = require ("express");
		global.formidable = require ("formidable");
		global.nodemailer = require ("nodemailer");
		global.simplesmtp = require ("simplesmtp");
		global.smtp = simplesmtp.createServer ();
		global.MailParser = require ("mailparser").MailParser;
		global.VError = require ("verror");
		global.Backbone = require ("backbone");
		try {
			config.log = config.log || {};
			config.log.level = config.log.level || "info";
			let bunyan = require ("bunyan");
			global.log = bunyan.createLogger ({
				name: "objectum",
				level: config.log.level || "info",
				streams: [{
					stream: process.stdout
				}, {
					path: config.rootDir + "/objectum-bunyan.log"
				}]
			});
		} catch (e) {
			let levelNum = {"fatal": 60, "error": 50, "warn": 40, "info": 30, "debug": 20, "trace": 10};
			config.log.levelNum = levelNum [config.log.level];
			function msgLog (level, opts, msg) {
				if (level < config.log.levelNum) {
					return;
				};
				opts = opts || {};
				if (typeof (opts) == "string") {
					opts = {msg: opts};
				};
				opts.msg = opts.msg || msg;
				opts.level = level;
				common.log ({file: config.rootDir + "/objectum-bunyan.log", text: util.inspect (opts, {depth: null})});
			};
			global.log = {
				fatal: function (opts, msg) {msgLog (60, opts, msg);},
				error: function (opts, msg) {msgLog (50, opts, msg);},
				warn: function (opts, msg) {msgLog (40, opts, msg);},
				info: function (opts, msg) {msgLog (30, opts, msg);},
				debug: function (opts, msg) {msgLog (20, opts, msg);},
				trace: function (opts, msg) {msgLog (10, opts, msg);}
			};
		};
		let sql;
		if (process.platform == "win32") {
			if (!config.mssql || config.mssql.enabled) {
				let tokens = process.version.split (".");
				if (tokens [1] <= 8 && tokens [2] <= 26) { // <= v0.8.26
					try {
						sql = require ("msnodesql");
					} catch (e) {
					}
				};
			};
		};
		let redisEmulator = {
			m: {},
			del: function (key, cb) {
				delete this.m [key];
				if (cb) {
					cb (null, true);
				};
			},
			keys: function (mask, cb) {
				let r = [];
				if (mask && mask [mask.length - 1] == "*") {
					let m = mask.substr (0, mask.length - 1);
					for (let key in this.m) {
						if (key.substr (0, m.length) == m) {
							r.push (key);
						};
					};
				};
				if (cb) {
					cb (null, r);
				};
			},
			subscribers: [],
			on: function (event, cb) {
				this.subscribers.push (cb);
			},
			subscribe: function (channel) {
			},
			publish: function (channel, message) {
				for (let i = 0; i < this.subscribers.length; i ++) {
					(this.subscribers [i]) (channel, message);
				};
			},
			incrby: function (key, inc, cb) {
				this.m [key] = this.m [key] || 0;
				this.m [key] = Number (this.m [key]) + inc;
				if (cb) {
					cb (null, String (this.m [key]));
				};
			},
			get: function (key, cb) {
				if (cb) {
					cb (null, this.m [key] || null);
				};
			},
			set: function (key, value, cb) {
				this.m [key] = String (value);
				if (cb) {
					cb (null, true);
				};
			},
			hincrby: function (key, field, inc, cb) {
				this.m [key] = this.m [key] || {};
				this.m [key][field] = this.m [key][field] || 0;
				this.m [key][field] = Number (this.m [key][field]) + inc;
				if (cb) {
					cb (null, String (this.m [key][field]));
				};
			},
			hdel: function () {
				let key = arguments [0];
				for (let i = 1; i < arguments.length; i ++) {
					if (typeof (arguments [i]) != "function" && this.m [key] && this.m [key][arguments [i]]) {
						delete this.m [key][arguments [i]];
					};
				};
				if (typeof (arguments [arguments.length - 1]) == "function") {
					(arguments [arguments.length - 1]) (null, true);
				};
			},
			hset: function (key, field, value, cb) {
				this.m [key] = this.m [key] || {};
				this.m [key][field] = String (value);
				if (cb) {
					cb (null, true);
				};
			},
			hmset: function (key, hdata, cb) {
				this.m [key] = this.m [key] || {};
				for (let field in hdata) {
					this.m [key][field] = String (hdata [field]);
				};
				if (cb) {
					cb (null, true);
				};
			},
			hget: function (key, field, cb) {
				this.m [key] = this.m [key] || {};
				if (cb) {
					cb (null, this.m [key][field]);
				};
			},
			hmget: function (key, keys, cb) {
				this.m [key] = this.m [key] || {};
				let r = [];
				for (let i = 0; i < keys.length; i ++) {
					r.push (this.m [key][keys [i]]);
				};
				if (cb) {
					cb (null, r);
				};
			},
			hgetall: function (key, cb) {
				let r = {};
				for (let field in this.m [key]) {
					r [field] = this.m [key][field];
				};
				if (cb) {
					cb (null, r);
				};
			},
			hsetnx: function (key, field, value, cb) {
				this.m [key] = this.m [key] || {};
				let r = 0;
				if (!this.m [key] [field]) {
					this.m [key] [field] = value;
					r = 1;
				};
				if (cb) {
					cb (null, r);
				};
			},
			hkeys: function (key, cb) {
				let r = [];
				for (let field in this.m [key]) {
					r.push (field);
				};
				if (cb) {
					cb (null, r);
				};
			},
			sadd: function (key, value, cb) {
				this.m [key] = this.m [key] || [];
				if (this.m [key].indexOf (value) == -1) {
					this.m [key].push (String (value));
					if (cb) {
						cb (null, true);
					};
				} else {
					if (cb) {
						cb (true, false);
					};
				};
			},
			sismember: function (key, value, cb) {
				this.m [key] = this.m [key] || [];
				if (cb) {
					cb (null, this.m [key].indexOf (value) > -1);
				};
			},
			srem: function (key, value, cb) {
				this.m [key] = this.m [key] || [];
				let pos = this.m [key].indexOf (String (value));
				if (pos > -1) {
					this.m [key].splice (pos, 1);
				};
				if (cb) {
					cb (null, true);
				};
			},
			smembers: function (key, cb) {
				this.m [key] = this.m [key] || [];
				if (cb) {
					cb (null, this.m [key]);
				};
			},
			lpush: function (key, value) {
				this.m [key] = this.m [key] || [];
				this.m [key] = [value].concat (this.m [key]);
				if (key.indexOf ("service.log") > -1) {
					console.log (key + " lpush " + this.m [key]);
				};
			},
			ltrim: function (key, start, end) {
				this.m [key] = this.m [key] || [];
				this.m [key] = this.m [key].slice (start, end + 1);
				if (key.indexOf ("service.log") > -1) {
					console.log (key + " ltrim " + this.m [key]);
				};
			},
			lrange: function (key, start, end, cb) {
				this.m [key] = this.m [key] || [];
				end = end == -1 ? (this.m [key].length - 1) : end;
				if (cb) {
					cb (null, this.m [key].slice (start, end + 1));
				};
				if (key.indexOf ("service.log") > -1) {
					console.log (key + " lrange " + this.m [key]);
				};
			},
			select: function (db, cb) {
				cb ();
			},
			quit: function () {
			}
		};
		if (!config.redis.enabled) {
			global.redis = {
				createClient: function () {
					return redisEmulator;
				}
			};
		};
		global.redisClient = redis.createClient (config.redis.port, config.redis.host, config.redis.password ? {password: config.redis.password} : undefined);
		global.redisPub = redis.createClient (config.redis.port, config.redis.host, config.redis.password ? {password: config.redis.password} : undefined);
		global.redisSub = redis.createClient (config.redis.port, config.redis.host, config.redis.password ? {password: config.redis.password} : undefined);

		require ("./db/client");
		require ("./db/mssql");
		require ("./db/postgres");
		require ("./report/dbf");
		require ("./report/pdf");
		require ("./report/xlsx");
		require ("./report/xmlss");
		require ("./common");
		require ("./export");
		require ("./import");
		require ("./mail");
		require ("./meta");
		require ("./mimetypes");
		require ("./projects");
		require ("./query");
		require ("./server");
		require ("./storage");
		require ("./toc");
		require ("./sha");
		require ("./ose");
		require ("./deprecated/projects");

		me.common = common;
		me.config = config;
		me.Export = Export;
		me.Import = Import;
		me.mailSender = mailSender;
		me.meta = meta;
		me.mimetypes = mimetypes;
		me.projects = projects;
		me.Query = Query;
		me.server = server;
		me.sha = sha;
		me.Storage = Storage;
		me.db = db;
		me.xmlss = xmlss;
		me.Dbf = Dbf;
		me.async = async;
		me.sha = sha;
		me.tm = tm;
		me.log = log;
		me.modules = {
			async: async,
			pg: pg,
			util: util,
			fs: fs,
			http: http,
			url: url,
			redis: redis,
			express: express,
			nodemailer: nodemailer,
			simplesmtp: simplesmtp,
			_: _
		};
	}
};
exports.Objectum = Objectum;
